#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import time
from typing import Any

from datadog import api, initialize  # type: ignore[attr-defined]

from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook
from airflow.utils.log.logging_mixin import LoggingMixin


class DatadogHook(BaseHook, LoggingMixin):
    """
    Uses datadog API to send metrics of practically anything measurable.

    It's possible to track # of db records inserted/deleted, records read
    from file and many other useful metrics.

    Depends on the datadog API, which has to be deployed on the same server where
    Airflow runs.

    :param datadog_conn_id: The connection to datadog, containing metadata for api keys.
    """

    conn_name_attr = "datadog_conn_id"
    default_conn_name = "datadog_default"
    conn_type = "datadog"
    hook_name = "Datadog"

    def __init__(self, datadog_conn_id: str = "datadog_default") -> None:
        super().__init__()
        conn = self.get_connection(datadog_conn_id)
        self.api_key = conn.extra_dejson.get("api_key", None)
        self.app_key = conn.extra_dejson.get("app_key", None)
        self.api_host = conn.extra_dejson.get("api_host", None)
        self.source_type_name = conn.extra_dejson.get("source_type_name", None)

        # If the host is populated, it will use that hostname instead.
        # for all metric submissions.
        self.host = conn.host

        if self.api_key is None:
            raise AirflowException("api_key must be specified in the Datadog connection details")

        self.log.info("Setting up api keys for Datadog")
        initialize(api_key=self.api_key, app_key=self.app_key, api_host=self.api_host)

    def validate_response(self, response: dict[str, Any]) -> None:
        """Validate Datadog response."""
        if response["status"] != "ok":
            self.log.error("Datadog returned: %s", response)
            raise AirflowException("Error status received from Datadog")

    def send_metric(
        self,
        metric_name: str,
        datapoint: float | int,
        tags: list[str] | None = None,
        type_: str | None = None,
        interval: int | None = None,
    ) -> dict[str, Any]:
        """
        Send a single datapoint metric to Datadog.

        :param metric_name: The name of the metric
        :param datapoint: A single integer or float related to the metric
        :param tags: A list of tags associated with the metric
        :param type_: Type of your metric: gauge, rate, or count
        :param interval: If the type of the metric is rate or count, define the corresponding interval
        """
        response = api.Metric.send(
            metric=metric_name, points=datapoint, host=self.host, tags=tags, type=type_, interval=interval
        )

        self.validate_response(response)
        return response

    def query_metric(self, query: str, from_seconds_ago: int, to_seconds_ago: int) -> dict[str, Any]:
        """
        Query datadog for a metric, potentially with some function applied to it and return the result.

        :param query: The datadog query to execute (see datadog docs)
        :param from_seconds_ago: How many seconds ago to start querying for.
        :param to_seconds_ago: Up to how many seconds ago to query for.
        """
        now = int(time.time())

        response = api.Metric.query(start=now - from_seconds_ago, end=now - to_seconds_ago, query=query)

        self.validate_response(response)
        return response

    def post_event(
        self,
        title: str,
        text: str,
        aggregation_key: str | None = None,
        alert_type: str | None = None,
        date_happened: int | None = None,
        handle: str | None = None,
        priority: str | None = None,
        related_event_id: int | None = None,
        tags: list[str] | None = None,
        device_name: list[str] | None = None,
    ) -> dict[str, Any]:
        """
        Post an event to datadog (processing finished, potentially alerts, other issues).

        Think about this as a means to maintain persistence of alerts, rather than alerting itself.

        :param title: The title of the event
        :param text: The body of the event (more information)
        :param aggregation_key: Key that can be used to aggregate this event in a stream
        :param alert_type: The alert type for the event, one of
            ["error", "warning", "info", "success"]
        :param date_happened: POSIX timestamp of the event; defaults to now
        :handle: User to post the event as; defaults to owner of the application key used
            to submit.
        :param handle: str
        :param priority: Priority to post the event as. ("normal" or "low", defaults to "normal")
        :param related_event_id: Post event as a child of the given event
        :param tags: List of tags to apply to the event
        :param device_name: device_name to post the event with
        """
        response = api.Event.create(
            title=title,
            text=text,
            aggregation_key=aggregation_key,
            alert_type=alert_type,
            date_happened=date_happened,
            handle=handle,
            priority=priority,
            related_event_id=related_event_id,
            tags=tags,
            host=self.host,
            device_name=device_name,
            source_type_name=self.source_type_name,
        )

        self.validate_response(response)
        return response

    @classmethod
    def get_connection_form_widgets(cls) -> dict[str, Any]:
        """Return connection widgets to add to connection form."""
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from flask_babel import lazy_gettext
        from wtforms import StringField

        return {
            "api_host": StringField(lazy_gettext("API endpoint"), widget=BS3TextFieldWidget()),
            "api_key": StringField(lazy_gettext("API key"), widget=BS3TextFieldWidget()),
            "app_key": StringField(lazy_gettext("Application key"), widget=BS3TextFieldWidget()),
            "source_type_name": StringField(lazy_gettext("Source type name"), widget=BS3TextFieldWidget()),
        }

    @classmethod
    def get_ui_field_behaviour(cls) -> dict[str, Any]:
        """Return custom field behaviour."""
        return {
            "hidden_fields": ["schema", "login", "password", "port", "extra"],
            "relabeling": {"host": "Events host name"},
        }



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from unittest.mock import patch

import pytest

from airflow.exceptions import AirflowException
from airflow.models import Connection
from airflow.providers.datadog.sensors.datadog import DatadogSensor
from airflow.utils import db

pytestmark = pytest.mark.db_test

at_least_one_event = [
    {
        "alert_type": "info",
        "comments": [],
        "date_happened": 1419436860,
        "device_name": None,
        "host": None,
        "id": 2603387619536318140,
        "is_aggregate": False,
        "priority": "normal",
        "resource": "/api/v1/events/2603387619536318140",
        "source": "My Apps",
        "tags": ["application:web", "version:1"],
        "text": "And let me tell you all about it here!",
        "title": "Something big happened!",
        "url": "/event/jump_to?event_id=2603387619536318140",
    },
    {
        "alert_type": "info",
        "comments": [],
        "date_happened": 1419436865,
        "device_name": None,
        "host": None,
        "id": 2603387619536318141,
        "is_aggregate": False,
        "priority": "normal",
        "resource": "/api/v1/events/2603387619536318141",
        "source": "My Apps",
        "tags": ["application:web", "version:1"],
        "text": "And let me tell you all about it here!",
        "title": "Something big happened!",
        "url": "/event/jump_to?event_id=2603387619536318141",
    },
]

zero_events: list = []


class TestDatadogSensor:
    def setup_method(self):
        db.merge_conn(
            Connection(
                conn_id="datadog_default",
                conn_type="datadog",
                login="login",
                password="password",
                extra=json.dumps({"api_key": "api_key", "app_key": "app_key"}),
            )
        )

    @patch("airflow.providers.datadog.hooks.datadog.api.Event.query")
    @patch("airflow.providers.datadog.sensors.datadog.api.Event.query")
    def test_sensor_ok(self, api1, api2):
        api1.return_value = at_least_one_event
        api2.return_value = at_least_one_event

        sensor = DatadogSensor(
            task_id="test_datadog",
            datadog_conn_id="datadog_default",
            from_seconds_ago=3600,
            up_to_seconds_from_now=0,
            priority=None,
            sources=None,
            tags=None,
            response_check=None,
        )

        assert sensor.poke({})

    @patch("airflow.providers.datadog.hooks.datadog.api.Event.query")
    @patch("airflow.providers.datadog.sensors.datadog.api.Event.query")
    def test_sensor_fail(self, api1, api2):
        api1.return_value = zero_events
        api2.return_value = zero_events

        sensor = DatadogSensor(
            task_id="test_datadog",
            datadog_conn_id="datadog_default",
            from_seconds_ago=0,
            up_to_seconds_from_now=0,
            priority=None,
            sources=None,
            tags=None,
            response_check=None,
        )

        assert not sensor.poke({})

    @patch("airflow.providers.datadog.hooks.datadog.api.Event.query")
    @patch("airflow.providers.datadog.sensors.datadog.api.Event.query")
    def test_sensor_fail_with_exception(self, api1, api2):
        api1.return_value = zero_events
        api2.return_value = {"status": "error"}

        sensor = DatadogSensor(
            task_id="test_datadog",
            datadog_conn_id="datadog_default",
            from_seconds_ago=0,
            up_to_seconds_from_now=0,
            priority=None,
            sources=None,
            tags=None,
            response_check=None,
        )
        with pytest.raises(AirflowException):
            sensor.poke({})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from unittest import mock

import pytest

from airflow.exceptions import AirflowException
from airflow.models import Connection
from airflow.providers.datadog.hooks.datadog import DatadogHook

APP_KEY = "app_key"
API_KEY = "api_key"
API_HOST = "api_host"
METRIC_NAME = "metric"
DATAPOINT = 7
TAGS = ["tag"]
TYPE = "rate"
INTERVAL = 30
TITLE = "title"
TEXT = "text"
AGGREGATION_KEY = "aggregation-key"
ALERT_TYPE = "warning"
DATE_HAPPENED = 12345
HANDLE = "handle"
PRIORITY = "normal"
RELATED_EVENT_ID = 7
DEVICE_NAME = "device-name"


class TestDatadogHook:
    def setup_method(self):
        with mock.patch("airflow.providers.datadog.hooks.datadog.initialize"):
            with mock.patch("airflow.providers.datadog.hooks.datadog.DatadogHook.get_connection") as m:
                m.return_value = Connection(
                    extra=json.dumps(
                        {
                            "app_key": APP_KEY,
                            "api_key": API_KEY,
                            "api_host": API_HOST,
                        }
                    )
                )
                self.hook = DatadogHook()

    @mock.patch("airflow.providers.datadog.hooks.datadog.initialize")
    @mock.patch("airflow.providers.datadog.hooks.datadog.DatadogHook.get_connection")
    def test_api_key_required(self, mock_get_connection, mock_initialize):
        mock_get_connection.return_value = Connection()
        with pytest.raises(AirflowException) as ctx:
            DatadogHook()
        assert str(ctx.value) == "api_key must be specified in the Datadog connection details"

    def test_validate_response_valid(self):
        try:
            self.hook.validate_response({"status": "ok"})
        except AirflowException:
            self.fail("Unexpected AirflowException raised")

    def test_validate_response_invalid(self):
        with pytest.raises(AirflowException):
            self.hook.validate_response({"status": "error"})

    @mock.patch("airflow.providers.datadog.hooks.datadog.api.Metric.send")
    def test_send_metric(self, mock_send):
        mock_send.return_value = {"status": "ok"}
        self.hook.send_metric(
            METRIC_NAME,
            DATAPOINT,
            tags=TAGS,
            type_=TYPE,
            interval=INTERVAL,
        )
        mock_send.assert_called_once_with(
            metric=METRIC_NAME,
            points=DATAPOINT,
            host=self.hook.host,
            tags=TAGS,
            type=TYPE,
            interval=INTERVAL,
        )

    @mock.patch("airflow.providers.datadog.hooks.datadog.api.Metric.query")
    @mock.patch("airflow.providers.datadog.hooks.datadog.time.time")
    def test_query_metric(self, mock_time, mock_query):
        now = 12345
        mock_time.return_value = now
        mock_query.return_value = {"status": "ok"}
        self.hook.query_metric("query", 60, 30)
        mock_query.assert_called_once_with(
            start=now - 60,
            end=now - 30,
            query="query",
        )

    @mock.patch("airflow.providers.datadog.hooks.datadog.api.Event.create")
    def test_post_event(self, mock_create):
        mock_create.return_value = {"status": "ok"}
        self.hook.post_event(
            TITLE,
            TEXT,
            aggregation_key=AGGREGATION_KEY,
            alert_type=ALERT_TYPE,
            date_happened=DATE_HAPPENED,
            handle=HANDLE,
            priority=PRIORITY,
            related_event_id=RELATED_EVENT_ID,
            tags=TAGS,
            device_name=DEVICE_NAME,
        )
        mock_create.assert_called_once_with(
            title=TITLE,
            text=TEXT,
            aggregation_key=AGGREGATION_KEY,
            alert_type=ALERT_TYPE,
            date_happened=DATE_HAPPENED,
            handle=HANDLE,
            priority=PRIORITY,
            related_event_id=RELATED_EVENT_ID,
            tags=TAGS,
            host=self.hook.host,
            device_name=DEVICE_NAME,
            source_type_name=self.hook.source_type_name,
        )
